import math

from delta.lexer.constants import NumericalConstants

from .function import Function
from delta.parser.nodes import Node, BinaryNode, UnaryNode
from delta.lexer.lexicalToken import LexicalToken
from delta.lexer.tokens import *


# d(log(a, u)) / dx = (1 / (log(a) * u)) * du / dx
def log_d(node):
    a = node.left_child
    u = node.right_child
    log_a = BinaryNode(LexicalToken(TT_FUNC, LOG), Node(LexicalToken(TT_REAL, NumericalConstants.get_constant('e'))), a)
    log_a_u = BinaryNode(LexicalToken(TT_MULTIPLY), log_a, u)
    _1_log_a_u = BinaryNode(LexicalToken(TT_DIVIDE), Node(LexicalToken(TT_INT, 1)), log_a_u)
    return BinaryNode(LexicalToken(TT_MULTIPLY), _1_log_a_u)


# d(sqrt(u)) / dx = (1 / 2 * sqrt(u)) * du / dx
def sqrt_d(node):
    u = node.right_child
    sqrt_u = UnaryNode(LexicalToken(TT_FUNC, SQRT), u)
    _2_sqrt_u = BinaryNode(LexicalToken(TT_MULTIPLY), Node(LexicalToken(TT_INT, 2)), sqrt_u)
    _1_2_sqrt_u = BinaryNode(LexicalToken(TT_DIVIDE), Node(LexicalToken(TT_INT, 1)), _2_sqrt_u)
    return BinaryNode(LexicalToken(TT_MULTIPLY), _1_2_sqrt_u)


# d(sin u) / dx = cos u * du / dx
def sin_d(node):
    u = node.right_child
    cos_u = UnaryNode(LexicalToken(TT_FUNC, COS), u)
    return BinaryNode(LexicalToken(TT_MULTIPLY), cos_u)


# d(cos u) / dx = (0 - sin u) * du / dx
def cos_d(node):
    u = node.right_child
    sin_u = UnaryNode(LexicalToken(TT_FUNC, SIN), u)
    _sin_u = BinaryNode(LexicalToken(TT_MINUS), Node(LexicalToken(TT_INT, 0)), sin_u)
    return BinaryNode(LexicalToken(TT_MULTIPLY), _sin_u)


# d(tan u) / dx =  sec ^ 2 (u) * du / dx
def tan_d(node):
    u = node.right_child
    sec_u = UnaryNode(LexicalToken(TT_FUNC, SEC), u)
    sec_2_u = BinaryNode(LexicalToken(TT_EXPONENT), sec_u, Node(LexicalToken(TT_INT, 2)))
    return BinaryNode(LexicalToken(TT_MULTIPLY), sec_2_u)


# d(sec u) / dx = sec u * tan u * du / dx
def sec_d(node):
    u = node.right_child
    sec_u = UnaryNode(LexicalToken(TT_FUNC, SEC), u)
    tan_u = UnaryNode(LexicalToken(TT_FUNC, TAN), u)
    sec_u_tan_u = BinaryNode(LexicalToken(TT_MULTIPLY), sec_u, tan_u)
    return BinaryNode(LexicalToken(TT_MULTIPLY), sec_u_tan_u)


# d(cosec u) / dx = (0 - cosec u) * cot u * du / dx
def cosec_d(node):
    u = node.right_child
    cosec_u = UnaryNode(LexicalToken(TT_FUNC, COSEC), u)
    _cosec_u = BinaryNode(LexicalToken(TT_MINUS), Node(LexicalToken(TT_INT, 0)), cosec_u)
    cot_u = UnaryNode(LexicalToken(TT_FUNC, COT), u)
    _cosec_u_cot_u = BinaryNode(LexicalToken(TT_MULTIPLY), _cosec_u, cot_u)
    return BinaryNode(LexicalToken(TT_MULTIPLY), _cosec_u_cot_u)


# d(cot u) / dx = ( 0 - cosec ^ 2 (u)) * du / dx
def cot_d(node):
    u = node.right_child
    cosec_u = UnaryNode(LexicalToken(TT_FUNC, COSEC), u)
    cosec_2_u = BinaryNode(LexicalToken(TT_EXPONENT), cosec_u, Node(LexicalToken(TT_INT, 2)))
    _cosec_2_u = BinaryNode(LexicalToken(TT_MINUS), Node(LexicalToken(TT_INT, 0)), cosec_2_u)
    return BinaryNode(LexicalToken(TT_MULTIPLY), _cosec_2_u)


# d(arcsin u) / dx = (1 / sqrt(1 - u ^ 2)) * du / dx
def arcsin_d(node):
    u = node.right_child
    u_2 = BinaryNode(LexicalToken(TT_EXPONENT), u, Node(LexicalToken(TT_INT, 2)))
    _1_u_2 = BinaryNode(LexicalToken(TT_MINUS), Node(LexicalToken(TT_INT, 1)), u_2)
    sqrt_v = UnaryNode(LexicalToken(TT_FUNC, SQRT), _1_u_2)
    _1_sqrt_v = BinaryNode(LexicalToken(TT_DIVIDE), Node(LexicalToken(TT_INT, 1)), sqrt_v)
    return BinaryNode(LexicalToken(TT_MULTIPLY), _1_sqrt_v)


# d(arccos u) / dx = (0 - (1 / sqrt(1 - u ^ 2))) * du / dx
def arccos_d(node):
    u = node.right_child
    u_2 = BinaryNode(LexicalToken(TT_EXPONENT), u, Node(LexicalToken(TT_INT, 2)))
    _1_u_2 = BinaryNode(LexicalToken(TT_MINUS), Node(LexicalToken(TT_INT, 1)), u_2)
    sqrt_v = UnaryNode(LexicalToken(TT_FUNC, SQRT), _1_u_2)
    _1_sqrt_v = BinaryNode(LexicalToken(TT_DIVIDE), Node(LexicalToken(TT_INT, 1)), sqrt_v)
    __1_sqrt_v = BinaryNode(LexicalToken(TT_MINUS), Node(LexicalToken(TT_INT, 0)), _1_sqrt_v)
    return BinaryNode(LexicalToken(TT_MULTIPLY), __1_sqrt_v)


# d(arctan u) / dx = (1 / (1 + u ^ 2)) * du / dx
def arctan_d(node):
    u = node.right_child
    u_2 = BinaryNode(LexicalToken(TT_EXPONENT), u, Node(LexicalToken(TT_INT, 2)))
    _1_u_2 = BinaryNode(LexicalToken(TT_PLUS), Node(LexicalToken(TT_INT, 1)), u_2)
    _1_v = BinaryNode(LexicalToken(TT_DIVIDE), Node(LexicalToken(TT_INT, 1)), _1_u_2)
    return BinaryNode(LexicalToken(TT_MULTIPLY), _1_v)


# d(arcsec u) / dx = (1 / u * sqrt(u ^ 2 - 1)) * du / dx
def arcsec_d(node):
    u = node.right_child
    u_2 = BinaryNode(LexicalToken(TT_EXPONENT), u, Node(LexicalToken(TT_INT, 2)))
    u_2_1 = BinaryNode(LexicalToken(TT_MINUS), u_2, Node(LexicalToken(TT_INT, 1)))
    u_u_2_1 = BinaryNode(LexicalToken(TT_MULTIPLY), u, u_2_1)
    _1_v = BinaryNode(LexicalToken(TT_DIVIDE), Node(LexicalToken(TT_INT, 1)), u_u_2_1)
    return BinaryNode(LexicalToken(TT_MULTIPLY), _1_v)


# d(arccosec u) / dx = (0 - (1 / u * sqrt(u ^ 2 - 1))) * du / dx
def arccosec_d(node):
    u = node.right_child
    u_2 = BinaryNode(LexicalToken(TT_EXPONENT), u, Node(LexicalToken(TT_INT, 2)))
    u_2_1 = BinaryNode(LexicalToken(TT_MINUS), u_2, Node(LexicalToken(TT_INT, 1)))
    u_u_2_1 = BinaryNode(LexicalToken(TT_MULTIPLY), u, u_2_1)
    _1_v = BinaryNode(LexicalToken(TT_DIVIDE), Node(LexicalToken(TT_INT, 1)), u_u_2_1)
    __1_v = BinaryNode(LexicalToken(TT_MINUS), Node(LexicalToken(TT_INT, 0)), _1_v)
    return BinaryNode(LexicalToken(TT_MULTIPLY), __1_v)


# d(arccot u) / dx = (0 - (1 / (1 + u ^ 2))) * du / dx
def arccot_d(node):
    u = node.right_child
    u_2 = BinaryNode(LexicalToken(TT_EXPONENT), Node(LexicalToken(TT_INT, 2)), u)
    _1_u_2 = BinaryNode(LexicalToken(TT_PLUS), Node(LexicalToken(TT_INT, 1)), u_2)
    _1_v = BinaryNode(LexicalToken(TT_DIVIDE), Node(LexicalToken(TT_INT, 1)), _1_u_2)
    __1_v = BinaryNode(LexicalToken(TT_MINUS), Node(LexicalToken(TT_INT, 0)), _1_v)
    return BinaryNode(LexicalToken(TT_MULTIPLY), __1_v)


def log_f(node):
    a = node.left_child
    b = node.right_child
    if a.lexical_token == b.lexical_token:
        return Node(LexicalToken(TT_INT, 1))

    return node


def sqrt_f(node):
    a = node.right_child
    if a.lexical_token == TT_INT or a.lexical_token == TT_REAL:
        val = a.lexical_token.value
        if a.lexical_token == LexicalToken(TT_REAL, 'e'):
            val = math.e
        return Node(LexicalToken(TT_REAL, pow(val, 0.5)))

    return node


LOG = Function('log', log_d, log_f, 1)

SQRT = Function('sqrt', sqrt_d, sqrt_f)

COS = Function('cos', cos_d)
SIN = Function('sin', sin_d)
TAN = Function('tan', tan_d)
SEC = Function('sec', sec_d)
COSEC = Function('cosec', cosec_d)
COT = Function('cot', cot_d)

ARCSIN = Function('arcsin', arcsin_d)
ARCCOS = Function('arccos', arccos_d)
ARCTAN = Function('arctan', arctan_d)
ARCSEC = Function('arcsec', arcsec_d)
ARCCOSEC = Function('arccosec', arccosec_d)
ARCCOT = Function('arccot', arccot_d)



from typing import Callable


class Function:
    functions = []

    def __init__(self, function_name: str, derivative_func: Callable, fold_func: Callable = lambda x: x, isbinary: int = 0):
        self.function_name = function_name
        self.derivative_func = derivative_func
        self.fold_func = fold_func
        self.isbinary = isbinary
        Function.functions.append(self)

    def derivative(self, node):
        return self.derivative_func(node)

    def fold(self, node):
        return self.fold_func(node)

    def __repr__(self):
        return f"{self.function_name}"

    @staticmethod
    def get_function(function_name: str):
        for function in Function.functions:
            if function.function_name == function_name:
                return function

        return None



class Error(Exception):
    def __init__(self, typ, start_pos, end_pos, details) -> None:
        self.typ = typ
        self.details = details
        self.start_pos = start_pos
        self.end_pos = end_pos
    
    def __repr__(self) -> str:
        return f"{self.typ} : {self.details} ({self.start_pos.idx}, {self.end_pos.idx})"


class InvalidIdentifier(Error):
    def __init__(self, start_pos, end_pos, details) -> None:
        super().__init__("Invalid Identifier", start_pos, end_pos, details)



import sys

from delta.parser.nodes import Node, BinaryNode, UnaryNode
from delta.parser.parser import Parser
from delta.lexer.lexer import Lexer
from delta.lexer.tokens import *
from delta.lexer.mathFunctions import *
from delta.lexer.constants import NumericalConstants
from delta.lexer.lexicalToken import LexicalToken


class Expression:
    def __init__(self, expression: str):
        self.__lexer = Lexer(expression)
        self.__parser = Parser(self.__lexer.make_tokens())
        self.head_node = self.__parser.make_nodes()
        self.expression = expression

    @classmethod
    def from_nodes(cls, head_node):
        expression = Expression.to_string(head_node)
        expr = cls(expression)
        expr.head_node = head_node
        return expr

    @staticmethod
    def __inorder_traversal(node, string):
        if node:
            if not(node.lexical_token == TT_INT or node.lexical_token == TT_SYMBOL):
                string.append('(')

            if isinstance(node, BinaryNode) and not(node.lexical_token == TT_FUNC):
                Expression.__inorder_traversal(node.left_child, string)

            string.append(str(node))

            if isinstance(node, BinaryNode) and node.lexical_token == TT_FUNC:
                Expression.__inorder_traversal(node.left_child, string)

            if isinstance(node, UnaryNode) or isinstance(node, BinaryNode):
                if node.lexical_token == TT_FUNC:
                    string.append('(')
                
                Expression.__inorder_traversal(node.right_child, string)

                if node.lexical_token == TT_FUNC:
                    string.append(')')

            if not (node.lexical_token == TT_INT or node.lexical_token == TT_SYMBOL):
                string.append(')')

    @staticmethod
    def to_string(node):
        string = []
        Expression.__inorder_traversal(node, string)
        return ' '.join(string)

    def __repr__(self):
        return Expression.to_string(self.head_node)

    @staticmethod
    def __fold_node(node):
        lex = node.lexical_token
        if lex == TT_EXPONENT:
            a = node.left_child
            b = node.right_child
            if a.lexical_token == LexicalToken(TT_INT, 1) or b.lexical_token == LexicalToken(TT_INT, 0):
                return Node(LexicalToken(TT_INT, 1))
            elif b.lexical_token == LexicalToken(TT_INT, 1):
                return a
            elif a.lexical_token == LexicalToken(TT_INT, 0):
                return Node(LexicalToken(TT_INT, 0))

        elif lex == TT_MULTIPLY:
            a = node.left_child
            b = node.right_child
            if a.lexical_token == LexicalToken(TT_INT, 1):
                return b
            elif b.lexical_token == LexicalToken(TT_INT, 1):
                return a
            elif a.lexical_token == LexicalToken(TT_INT, 0) or b.lexical_token == LexicalToken(TT_INT, 0):
                return Node(LexicalToken(TT_INT, 0))

        elif lex == TT_PLUS:
            a = node.left_child
            b = node.right_child
            if a.lexical_token == LexicalToken(TT_INT, 0):
                return b
            elif b.lexical_token == LexicalToken(TT_INT, 0):
                return a

        elif lex == TT_MINUS:
            a = node.left_child
            b = node.right_child
            if b.lexical_token == LexicalToken(TT_INT, 0):
                return a

        elif lex == TT_DIVIDE:
            a = node.left_child
            b = node.right_child
            if b.lexical_token == LexicalToken(TT_INT, 1):
                return a

        elif lex == TT_FUNC:
            return node.lexical_token.value.fold(node)

        return node

    @staticmethod
    def __fold(node, prev_node, left):
        if node:

            if isinstance(node, BinaryNode):
                Expression.__fold(node.left_child, node, 1)

            if prev_node and node:
                if left:
                    prev_node.left_child = Expression.__fold_node(node)
                else:
                    prev_node.right_child = Expression.__fold_node(node)

            if isinstance(node, UnaryNode) or isinstance(node, BinaryNode):
                Expression.__fold(node.right_child, node, 0)

            if isinstance(node, UnaryNode) or isinstance(node, BinaryNode):
                prev_node = node

        return node

    @staticmethod
    def fold(node):
        prev_node = node
        for _ in range(5):
            node = Expression.__fold(prev_node, None, 0)
            node = Expression.__fold_node(node)
            prev_node = node
        return prev_node

    @staticmethod
    def __differentiate(node, wrt):

        lex = node.lexical_token

        # d(u + v) / dx = du / dx + dv / dx
        if lex == TT_PLUS:
            u = node.left_child
            v = node.right_child
            du = Expression.__differentiate(u, wrt)
            dv = Expression.__differentiate(v, wrt)
            return BinaryNode(LexicalToken(TT_PLUS), du, dv)

        # d(u - v) / dx = du / dx - dv / dx
        elif lex == TT_MINUS:
            u = node.left_child
            v = node.right_child
            du = Expression.__differentiate(u, wrt)
            dv = Expression.__differentiate(v, wrt)
            return BinaryNode(LexicalToken(TT_MINUS), du, dv)

        # d(uv) / dx = v * du / dx + u * dv / dx
        elif lex == TT_MULTIPLY:
            u = node.left_child
            v = node.right_child

            if u.lexical_token == TT_INT:
                return BinaryNode(LexicalToken(TT_MULTIPLY), u, Expression.__differentiate(v, wrt))

            if v.lexical_token == TT_INT:
                return BinaryNode(LexicalToken(TT_MULTIPLY), v, Expression.__differentiate(u, wrt))

            v_du = BinaryNode(LexicalToken(TT_MULTIPLY), v, Expression.__differentiate(u, wrt))
            u_dv = BinaryNode(LexicalToken(TT_MULTIPLY), u, Expression.__differentiate(v, wrt))
            return BinaryNode(LexicalToken(TT_PLUS), v_du, u_dv)

        # d(u / v) / dx = v * du - u * dv / v ^ 2
        elif lex == TT_DIVIDE:
            u = node.left_child
            v = node.right_child
            v_du = BinaryNode(LexicalToken(TT_MULTIPLY), v, Expression.__differentiate(u, wrt))
            u_dv = BinaryNode(LexicalToken(TT_MULTIPLY), u, Expression.__differentiate(v, wrt))
            numerator = BinaryNode(LexicalToken(TT_MINUS), v_du, u_dv)
            denominator = BinaryNode(LexicalToken(TT_EXPONENT), v, Node(LexicalToken(TT_INT, 2)))
            return BinaryNode(LexicalToken(TT_DIVIDE), numerator, denominator)

        # d(u ^ c) / dx =  c * u ^ (c - 1) * du / dx
        elif lex == TT_EXPONENT:
            u = node.left_child
            c = node.right_child

            # d(a ^ u) / dx = a ^ u * log(a) * du / dx
            if u.lexical_token == TT_INT or u.lexical_token == TT_REAL:
                a_u = BinaryNode(LexicalToken(TT_EXPONENT), u, c)
                log_a = BinaryNode(LexicalToken(TT_FUNC, LOG), Node(LexicalToken(TT_REAL, NumericalConstants.get_constant('e'))), u)
                a_u_log_a = BinaryNode(LexicalToken(TT_MULTIPLY), a_u, log_a)
                du = Expression.__differentiate(c, wrt)
                return BinaryNode(LexicalToken(TT_MULTIPLY), a_u_log_a, du)

            # d(u ^ c) / dx =  c * u ^ (c - 1) * du / dx
            elif c.lexical_token == TT_INT or c.lexical_token == TT_REAL or c.lexical_token == TT_SYMBOL:
                if c.lexical_token == wrt:
                    return node

                if c.lexical_token == TT_SYMBOL:
                    u_c_1 = BinaryNode(LexicalToken(TT_EXPONENT), u, Node(LexicalToken(TT_INT, str(c.lexical_token.value) + "-1")))
                else:
                    u_c_1 = BinaryNode(LexicalToken(TT_EXPONENT), u, Node(LexicalToken(TT_INT, c.lexical_token.value - 1)))

                c_u_c_1 = BinaryNode(LexicalToken(TT_MULTIPLY), c, u_c_1)
                du = Expression.__differentiate(u, wrt)
                return BinaryNode(LexicalToken(TT_MULTIPLY), c_u_c_1, du)

            else:
                sys.exit(0)

        elif lex == TT_FUNC:
            du = Expression.__differentiate(node.right_child, wrt)
            res = lex.value.derivative(node)
            res.right_child = du
            return res

        # d(c) / dx = 0
        elif lex in [TT_INT, TT_REAL]:
            return Node(LexicalToken(TT_INT, 0))

        # d(v) / dx = 1 (v : variable)
        elif lex == TT_SYMBOL:
            # d(x) / dx = 1
            if lex == wrt:
                return Node(LexicalToken(TT_INT, 1))

            # d(v) / dx = 0 (v : variable other than x)
            else:
                return Node(LexicalToken(TT_INT, 0))

    @staticmethod
    def differentiate(expression: "Expression", wrt: LexicalToken = LexicalToken(TT_SYMBOL, 'x'),
                      nth_derivative: int = 1, fold: bool = True):
        head = expression.head_node
        for _ in range(nth_derivative):
            head = Expression.__differentiate(head, wrt)
            if fold:
                head = Expression.fold(head)

        return Expression.from_nodes(head)



from typing import List, Text
from delta.lexer.mathFunctions import LOG

from delta.lexer.tokens import *
from delta.lexer.lexicalToken import LexicalToken
from .errors import InvalidExpression
from .nodes import BinaryNode, UnaryNode, Node

class Parser:

    precedence = {
            TT_PLUS : 1,
            TT_MINUS : 1,
            TT_MULTIPLY : 2,
            TT_DIVIDE : 3,
            TT_EXPONENT : 4,
            TT_FUNC : 5,
        }
    
    binary_nodes = [TT_PLUS, TT_MINUS, TT_DIVIDE, TT_MULTIPLY, TT_EXPONENT]
    unary_nodes = [TT_FUNC]
    nodes = [TT_REAL, TT_INT, TT_SYMBOL]

    def __init__(self, lexical_tokens: list):
        self.tokens = lexical_tokens
        self.head = None

    @staticmethod
    def __balanceOfBrackets(tokens):
        stack = []
        for i in range(len(tokens)):
            if tokens[i] == TT_LPAREN:
                stack.append((tokens[i], i))
            elif tokens[i] == TT_RPAREN:
                if stack and stack[-1][0] == TT_LPAREN:
                    stack.pop()
                else:
                    return False, (tokens[i], i)

        return (True, (None, None)) if not(stack) else (False, stack[-1])

    @staticmethod
    def __postfix(tokens: List[LexicalToken]):
        postfix_expr = []
        stack = []
        for token in tokens:
            if token in [TT_SYMBOL, TT_INT, TT_REAL]:
                postfix_expr.append(token)
            elif token in [TT_LPAREN]:
                stack.append(token)
            elif token in [TT_RPAREN]:
                while stack and stack[-1] != TT_LPAREN:
                    postfix_expr.append(stack.pop())

                if stack:
                    stack.pop()

            elif token in [TT_PLUS, TT_MINUS, TT_MULTIPLY, TT_DIVIDE, TT_EXPONENT, TT_FUNC]:
                while stack and Parser.precedence.get(token.token_type, 0) <= Parser.precedence.get(stack[-1].token_type, 0):
                    postfix_expr.append(stack.pop())
                stack.append(token)

        while stack:
            postfix_expr.append(stack.pop())

        return postfix_expr

    @staticmethod
    def __prefix(tokens: List[LexicalToken]):
        balanced, error = Parser.__balanceOfBrackets(tokens)
        if balanced:
            tokens = tokens[::-1]
            for i in range(len(tokens)):
                if tokens[i] == TT_LPAREN:
                    tokens[i] = LexicalToken(TT_RPAREN)
                elif tokens[i] == TT_RPAREN:
                    tokens[i] = LexicalToken(TT_LPAREN)

            return Parser.__postfix(tokens)[::-1]
        raise Exception(InvalidExpression(f"Check for balance of brackets : {error}"))

    @staticmethod
    def __make_nodes(tokens):
        elem = tokens.pop(0)
        if elem in Parser.nodes:
            return Node(elem)
        else:
            if elem in Parser.binary_nodes or (elem == TT_FUNC and elem.value.isbinary):
                left = Parser.__make_nodes(tokens)

            if elem in Parser.unary_nodes or elem in Parser.binary_nodes:
                right = Parser.__make_nodes(tokens)

        if elem in Parser.binary_nodes or (elem == TT_FUNC and elem.value.isbinary):
            return BinaryNode(elem, left, right)

        if elem in Parser.unary_nodes:
            return UnaryNode(elem, right)

    def prefix(self):
        return Parser.__prefix(self.tokens)

    def make_nodes(self):
        prefix_expr = Parser.__prefix(self.tokens)
        self.head = Parser.__make_nodes(prefix_expr)
        return self.head



from .tokenType import TokenType

# Variables, Numerals
TT_INT = TokenType('INT')
TT_REAL = TokenType('REAL')

# Special Variables
TT_SYMBOL = TokenType('SYMBOL')

# Special Mathematical Functions
TT_FUNC = TokenType('FUNCTION')

# Mathematical Operators
TT_PLUS = TokenType('PLUS', '+')
TT_MINUS = TokenType('MINUS', '-')
TT_DIVIDE = TokenType('DIVIDE', '/')
TT_MULTIPLY = TokenType('MULTIPLY', '*')
TT_EXPONENT = TokenType('EXPONENT', '^')

# Mathematical Expression Operators
TT_LPAREN = TokenType('LPAREN', '(')
TT_RPAREN = TokenType('RPAREN', ')')


from delta.lexer.lexicalToken import LexicalToken


class Node:
    def __init__(self, lexical_token: LexicalToken):
        self.lexical_token = lexical_token

    def __str__(self):
        return str(self.lexical_token)

    def __repr__(self):
        return f"{self.lexical_token}"


class BinaryNode(Node):
    def __init__(self, lexical_token: LexicalToken, left_child: Node = None, right_child: Node = None):
        super().__init__(lexical_token)
        self.left_child = left_child
        self.right_child = right_child


class UnaryNode(Node):
    def __init__(self, lexical_token: LexicalToken, right_child: Node = None):
        super().__init__(lexical_token)
        self.right_child = right_child



from typing import Optional

from delta.lexer.function import Function
from delta.lexer.mathFunctions import *
from .constants import NumericalConstants, SymbolicConstants

from .position import Position
from .tokens import *
from .lexicalToken import LexicalToken
from .errors import InvalidIdentifier


class Lexer:
    def __init__(self, expression: str):
        self.expression = expression
        self.tokens = []
        self.pos: Position = Position(-1)
        self.curr_char: Optional[str] = None
        self.advance()

    def advance(self):
        self.pos.advance()
        if self.pos.idx >= len(self.expression):
            self.curr_char = None
        else:
            self.curr_char = self.expression[self.pos.idx]

    def make_tokens(self):
        while self.curr_char:

            if self.curr_char in ' \n\t':
                self.advance()

            # Operators
            elif self.curr_char == '+':
                self.tokens.append(LexicalToken(TT_PLUS))
                self.advance()
            elif self.curr_char == '-':
                self.tokens.append(LexicalToken(TT_MINUS))
                self.advance()
            elif self.curr_char == '*':
                self.tokens.append(LexicalToken(TT_MULTIPLY))
                self.advance()
            elif self.curr_char == '/':
                self.tokens.append(LexicalToken(TT_DIVIDE))
                self.advance()
            elif self.curr_char == '^':
                self.tokens.append(LexicalToken(TT_EXPONENT))
                self.advance()

            # Parenthesis
            elif self.curr_char == ')':
                self.tokens.append(LexicalToken(TT_RPAREN))
                self.advance()
            elif self.curr_char == '(':
                self.tokens.append(LexicalToken(TT_LPAREN))
                self.advance()

            # Numerical Constants, Symbolic Constants, Identifiers
            elif self.curr_char.isdigit():
                self.tokens.append(self.make_digits())
            elif self.curr_char.isalpha():
                start_pos = Position.copy(self.pos)
                ident = self.make_identifiers()
                if isinstance(ident, LexicalToken):
                    self.tokens.append(ident)
                else:
                    raise InvalidIdentifier(start_pos, self.pos, ident )

            else:
                start_pos = Position.copy(self.pos)
                char = self.curr_char
                self.advance()
                raise InvalidIdentifier(start_pos, self.pos, char)

        return self.tokens

    def make_identifiers(self):
        identifier = ""
        while self.curr_char and self.curr_char.isalpha():
            identifier += self.curr_char
            self.advance()

        constant = NumericalConstants.get_constant(identifier)
        if constant:
            return LexicalToken(TT_REAL, constant)

        constant = SymbolicConstants.get_constant(identifier)
        if constant:
            return LexicalToken(TT_SYMBOL, constant)

        func = Function.get_function(identifier)
        if func:
            return LexicalToken(TT_FUNC, func)

        return identifier

    def make_digits(self):
        number = ""
        decimal = 0
        while self.curr_char and (self.curr_char.isdigit() or self.curr_char == '.'):
            if self.curr_char == '.':
                if decimal == 1:
                    break

                decimal += 1

            number += self.curr_char
            self.advance()

        if decimal > 0:
            return LexicalToken(TT_REAL, float(number))
        else:
            return LexicalToken(TT_INT, int(number))



class Position:
    def __init__(self, idx):
        self.idx = idx

    def advance(self):
        self.idx += 1

    @staticmethod
    def copy(other):
        if isinstance(other, Position):
            return Position(other.idx)
        return None



from typing import Any
from .tokenType import TokenType


class LexicalToken:
    def __init__(self, token_type: TokenType, value: Any = None):
        self.token_type = token_type
        self.value = value
        self.token_name = token_type.token_name
        self.token_symbol = token_type.token_symbol

    def __eq__(self, other):
        if isinstance(other, LexicalToken):
            return other.token_type == self.token_type and self.value == other.value
        elif isinstance(other, TokenType):
            return self.token_type == other
        else:
            return False

    def __str__(self):
        return str(self.value) if self.value is not None else str(self.token_type)

    def __repr__(self):
        return str(self.value) if self.value is not None else str(self.token_type)



from typing import Any


class TokenType:
    def __init__(self, token_name: str, token_symbol: Any = None):
        self.token_name = token_name
        self.token_symbol = token_symbol

    def __hash__(self) -> int:
        return self.token_symbol.__hash__()

    def __eq__(self, other: "TokenType"):
        return other.token_name == self.token_name and other.token_symbol == self.token_symbol

    def __str__(self):
        return str(self.token_symbol)

    def __repr__(self):
        return str(self.token_symbol)



import math


class NumericalConstants:
    constants = []

    def __init__(self, symbol, value):
        self.symbol = symbol
        self.value = value
        NumericalConstants.constants.append(self)

    def __repr__(self) -> str:
        return f"{self.symbol}"

    def __eq__(self, other):
        if isinstance(other, str):
            return self.symbol == other
        if isinstance(other, NumericalConstants):
            return self.symbol == other.symbol
            
        return False

    @staticmethod
    def get_constant(symbol):
        for constant in NumericalConstants.constants:
            if constant == symbol:
                return constant
        return None


class SymbolicConstants:
    constants = []

    def __init__(self, symbol):
        self.symbol = symbol
        SymbolicConstants.constants.append(self)

    def __repr__(self) -> str:
        return f"{self.symbol}"

    def __eq__(self, other):
        if isinstance(other, str):
            return self.symbol == other
        if isinstance(other, SymbolicConstants):
            return self.symbol == other.symbol
            
        return False

    @staticmethod
    def get_constant(symbol):
        for constant in SymbolicConstants.constants:
            if constant == symbol:
                return constant
        return None


e = NumericalConstants('e', math.e)
pi = NumericalConstants('pi', math.pi)
tau = NumericalConstants('ta', math.tau)

x = SymbolicConstants('x')
y = SymbolicConstants('y')
z = SymbolicConstants('z')
w = SymbolicConstants('w')
v = SymbolicConstants('v')
u = SymbolicConstants('u')
n = SymbolicConstants('n')



class Error:
    def __init__(self, typ, details) -> None:
        self.typ = typ
        self.details = details

    def __repr__(self) -> str:
        return f"{self.typ} : {self.details}"


class InvalidExpression(Error):
    def __init__(self, details) -> None:
        super().__init__("Invalid Expression", details)
